// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.inf.usi.ch/carzaniga/siena/
//
//  Authors: Antonio Carzaniga
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2003 University of Colorado
//
//  Siena is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//  
//  Siena is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with Siena.  If not, see <http://www.gnu.org/licenses/>.
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <siena/types.h>
#include <siena/forwarding.h>
#include <siena/allocator.h>
#include <siena/bctable.h>

#include "b_table.h"
#include "bitvector.h"
#include "counters_map.h"
#include "mmhwrapper.h"
#include "hash.h"
#include "bset.h"
#include "bset_encoding.h"

#include "timers.h"

namespace siena {

struct BCInterface {
    const if_t interface;
    const ifid_t id;

    /** builds an interface descriptor with the given identifiers **/
    BCInterface(if_t xif, ifid_t xid) : interface(xif), id(xid) {};
};

struct BCFilter {
    BCFilter(BCInterface * xi, unsigned int xid) 
	: i(xi), size(0), id(xid) {};

    BCInterface * i;
    unsigned int size;
    /*  filter identifier used as a key in the table of counters in
     *  the matching algorithm
     */
    unsigned int id;
};

struct BCFilterList {
    BCFilterList * next;
    BCFilter * f;

    BCFilterList(BCFilterList * nxt, BCFilter * fi) : next(nxt), f(fi) {};
};

BCTable::BCTable() : f_counter(0), if_counter(0) {
    for(unsigned int i = 0; i < BSize; ++i)
	positions[i] = 0;
}

BCTable::~BCTable() { 
    memory.clear();
}

void BCTable::consolidate() { 

    TIMER_PUSH(consolidate_timer);

    for(const BPredicate * p = plist; p; p = p->next) {
	BCInterface * I = new (memory) BCInterface(p->id, if_counter++);

	for(const BFilter * f = p->flist; f; f = f->next) {
	    BCFilter * bcf = new (memory)BCFilter(I, f_counter++);

	    for(unsigned int i = 0; i < BSize; ++i) {
		if (f->b[i]) {
		    positions[i] = new (memory)BCFilterList(positions[i], bcf);
		    ++(bcf->size);
		}
	    }
	}
    }
    TIMER_POP();
}

class BCProcessor {

public:
    BCProcessor(unsigned int if_count, 
		MatchHandler & h, 
		unsigned int max_pos,
		BCFilterList * const * parray)
	: fmap(), 
	  pos_mask(max_pos),
	  if_mask(if_count), 
	  handler(h), 
	  target(if_count),
	  positions(parray),
	  bsize(max_pos),
	  complete(false)
    { };

    unsigned int size() { return bsize; }

    bool set_bit(unsigned int pos);

    bool matching_complete() { return complete; }

private:
    /** table of counters for partially-matched filters.
     *
     *  Each filter considered in the matching process has an
     *  associated counter that records the number of constraints
     *  matched so far for that filter.
     */
    counters_map		fmap;

    bitvector 			pos_mask;

    /** set of interfaces that can be ignored in the matching process.
     *
     *  An interface can be ignored by the matching process if it has
     *  already been matched and processed or if it was excluded by
     *  the pre-processing function.
     */
    bitvector 			if_mask;
    /** output processor
     */
    MatchHandler &		handler;
    /** total number of interfaces we can match.
     *
     *  we maintain this value so that we can stop the matching
     *  process immediately once we have matched all interfaces.
     */
    unsigned int	target;

    BCFilterList * const * positions;

    unsigned int	bsize;

    bool		complete;
};

bool BCProcessor::set_bit(unsigned int pos) {
    if (pos_mask.set(pos)) 
	return false;

    for(const BCFilterList * l = positions[pos]; l; l = l->next) {
	// for each filter where the pos-th position is 1
	//
	if (! if_mask.test(l->f->i->id)) {
	    // if the interface to which the filter is associated has
	    // not been matched already
	    //
	    if (fmap.plus_one(l->f->id) >= l->f->size) {
		// if the filter has been matched, then mark this
		// interface as "matched", and call the processor on
		// it.  interface id into the result.
		//
		if_mask.set(l->f->i->id); 

		TIMER_PUSH(forward_timer);

		bool output_result = handler.output(l->f->i->interface);

		TIMER_POP();

		if (output_result) 
		    return complete = true;

		if (--target == 0) 
		    return complete = true;
	    }
	}
    }
    return false;
}

void BCTable::match(const message & m, MatchHandler & h) const {
    message::iterator * a  = m.first();
    if (!a) return;

    TIMER_PUSH(match_timer);

    BCProcessor p(if_counter, h, BSize, positions);

    do {
	encode_attribute(p, a);
    } while (!p.matching_complete() && a->next());
    delete(a);

    TIMER_POP();
}

} // end namespace siena

