// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.inf.usi.ch/carzaniga/siena/
//
//  Authors: Antonio Carzaniga
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2003 University of Colorado
//  Copyright (C) 2005 Antonio Carzaniga
//
//  Siena is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  (at your option) any later version.
//  
//  Siena is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with Siena.  If not, see <http://www.gnu.org/licenses/>.
//
#ifndef SIENA_BDDBTABLE_H
#define SIENA_BDDBTABLE_H

#include <cstddef>

#include <siena/forwarding.h>
#include <siena/allocator.h>

namespace siena {

/** @brief implementation of the forwarding table based on Bloom
 *         filters and BDDs. This is the common structure of the
 *         BDD-based algorithm.
 *
 *  The idea behind this algorithm is to represent each encoded
 *  predicate as a BDD.  An encoded predicate is constructed as an
 *  M-by-N bit matrix, where M is the chosen size of the Bloom filter,
 *  and N is the number of Bloom filters, each one representing a
 *  conjunction of constraints.  Therefore, each one of the M bits of
 *  the Bloom filters becomes a variable in the BDD (or ZDD).
 **/
class BDDBTableBase : public ForwardingTable {
public:
    virtual void consolidate();
    virtual void ifconfig(if_t, const predicate &) = 0;
    virtual void match(const message &, MatchHandler &) const = 0;
    virtual void match(const message & m, MatchMessageHandler & p) const;
    virtual size_t allocated_bytesize() const;
    virtual size_t bytesize() const;
    virtual void clear() = 0;
    virtual void clear_recycle() = 0;

protected:
    /** @brief Protected allocator of the forwarding table.  
     *
     *  All the data structure forming the forwarding table are
     *  allocated through this memory management system.
     **/
    FTAllocator memory;
};

/** @brief implementation of the forwarding table based on Bloom
 *         filters. This implementation consolidates each predicate
 *         into a BDD, which is then used for matching.
 *
 *  @see BDDBTableBase
 **/
class BDDBTable : public BDDBTableBase {
public:
    BDDBTable();
    virtual ~BDDBTable();

    virtual void ifconfig(if_t, const predicate &);
    virtual void match(const message &, MatchHandler &) const;

    virtual void clear();
    virtual void clear_recycle();

private:
    class IFDescr;

    IFDescr * bddlist;
};

/** @brief implementation of the forwarding table based on Bloom
 *         filters. This implementation consolidates each predicate
 *         into a ZDD, which is then used for matching.
 *
 *  @see BDDBTableBase
 **/
class ZDDBTable : public BDDBTableBase {
public:
    ZDDBTable();
    virtual ~ZDDBTable();

    virtual void ifconfig(if_t, const predicate &);
    virtual void match(const message &, MatchHandler &) const;

    virtual void clear();
    virtual void clear_recycle();

private:
    class IFDescr;

    IFDescr * zddlist;
};

}; // end namespace siena

#endif
